(* // Code by SoundOfSpouting#6980 (UID: 151149148639330304) *)

(* The constructor is `Entry` instead of `Entry()`, otherwise as required *)

type Color = Red | Black

type Entry<'Key, 'Value when 'Key: comparison> =
    | Empty
    | Tree of smaller: Entry<'Key, 'Value> * key: 'Key * value: 'Value
            * bigger: Entry<'Key, 'Value> * color: Color

    member this.TryFind (key: 'Key) : 'Value option =
        match this with
        | Empty -> None
        | Tree (s, k, v, b, _) ->
            if k = key then
                Some v
            else
                if key > k then b.TryFind key
                else s.TryFind key
    
    member this.Add (key : 'Key) (value: 'Value) : Entry<'Key, 'Value> =
        match this.ChildAdd key value with
        | Tree (s, k, v, b, Red) -> Tree (s, k, v, b, Black)
        | other -> other

    member private this.ChildAdd (key : 'Key) (value: 'Value) : Entry<'Key, 'Value> =
        match this with
        | Empty -> Tree (Empty, key, value, Empty, Red)
        | Tree (s, k, v, b, c) ->
            if k = key then Tree (s, key, value, b, c) else
            ( if key > k then Tree (s, k, v, b.ChildAdd key value, c)
              else Tree (s.ChildAdd key value, k, v, b, c) ).rebalance
    
    member private this.rebalance : Entry<'Key, 'Value> =
        match this with
        | Tree (Tree (Tree (a, x, xv, b, Red), y, yv, c, Red), z, zv, d, Black)
        | Tree (Tree (a, x, xv, Tree (b, y, yv, c, Red), Red), z, zv, d, Black)
        | Tree (a, x, xv, Tree (Tree (b, y, yv, c, Red), z, zv, d, Red), Black)
        | Tree (a, x, xv, Tree (b, y, yv, Tree (c, z, zv, d, Red), Red), Black) ->
            Tree (Tree (a, x, xv, b, Black), y, yv, Tree(c, z, zv, d, Black), Red)
        | other -> other